import numpy as np
import scipy as sp
from scipy.spatial.distance import pdist
from numba import jit
import sklearn
import warnings
from ._general import safe_isinstance


def partition_tree(X, metric="correlation"):
    X_full_rank = X + np.random.randn(*X.shape) * 1e-8
    D = sp.spatial.distance.pdist(X_full_rank.fillna(X_full_rank.mean()).T, metric=metric)
    return sp.cluster.hierarchy.complete(D)


def partition_tree_shuffle(indexes, index_mask, partition_tree):
    """ Randomly shuffle the indexes in a way that is consistent with the given partition tree.

    Parameters
    ----------
    indexes: np.array
        The output location of the indexes we want shuffled. Note that len(indexes) should equal index_mask.sum().

    index_mask: np.array
        A bool mask of which indexes we want to include in the shuffled list.

    partition_tree: np.array
        The partition tree we should follow.
    """
    M = len(index_mask)
    #switch = np.random.randn(M) < 0
    _pt_shuffle_rec(partition_tree.shape[0]-1, indexes, index_mask, partition_tree, M, 0)
@jit
def _pt_shuffle_rec(i, indexes, index_mask, partition_tree, M, pos):
    if i < 0:
        # see if we should include this index in the ordering
        if index_mask[i + M]: 
            indexes[pos] = i + M
            return pos + 1
        else:
            return pos
    left = int(partition_tree[i,0] - M)
    right = int(partition_tree[i,1] - M)
    if np.random.randn() < 0:
        pos = _pt_shuffle_rec(left, indexes, index_mask, partition_tree, M, pos)
        pos = _pt_shuffle_rec(right, indexes, index_mask, partition_tree, M, pos)
    else:
        pos = _pt_shuffle_rec(right, indexes, index_mask, partition_tree, M, pos)
        pos = _pt_shuffle_rec(left, indexes, index_mask, partition_tree, M, pos)
    return pos

@jit
def delta_minimization_order(all_masks, max_swap_size=100, num_passes=2):
    order = np.arange(len(all_masks))
    for _ in range(num_passes):
        for length in list(range(2, max_swap_size)): 
            for i in range(1, len(order)-length):
                if _reverse_window_score_gain(all_masks, order, i, length) > 0:
                    _reverse_window(order, i, length)
    return order
@jit
def _reverse_window(order, start, length):
    for i in range(length // 2):
        tmp = order[start + i]
        order[start + i] = order[start + length - i - 1]
        order[start + length - i - 1] = tmp
@jit
def _reverse_window_score_gain(masks, order, start, length):
    forward_score = _mask_delta_score(masks[order[start - 1]], masks[order[start]]) + \
                    _mask_delta_score(masks[order[start + length-1]], masks[order[start + length]])
    reverse_score = _mask_delta_score(masks[order[start - 1]], masks[order[start + length-1]]) + \
                    _mask_delta_score(masks[order[start]], masks[order[start + length]])
    
    return forward_score - reverse_score
@jit
def _mask_delta_score(m1, m2):
    return (m1 ^ m2).sum()


def hclust_ordering(X, metric="sqeuclidean", anchor_first=False):
    """ A leaf ordering is under-defined, this picks the ordering that keeps nearby samples similar.
    """
    
    # compute a hierarchical clustering and return the optimal leaf ordering
    D = sp.spatial.distance.pdist(X, metric)
    cluster_matrix = sp.cluster.hierarchy.complete(D)
    return sp.cluster.hierarchy.leaves_list(sp.cluster.hierarchy.optimal_leaf_ordering(cluster_matrix, D))
    
    # merge clusters, rotating them to make the end points match as best we can
    # sets = [[i] for i in range(X.shape[0])]
    # for i in range(cluster_matrix.shape[0]):
    #     s1 = sets[int(cluster_matrix[i,0])]
    #     s2 = sets[int(cluster_matrix[i,1])]

    #     # compute distances between the end points of the lists
    #     d_s1_s2 = pdist(np.vstack([X[s1[-1],:], X[s2[0],:]]), metric)[0]
    #     d_s1_s2r = pdist(np.vstack([X[s1[-1],:], X[s2[-1],:]]), metric)[0]
    #     d_s1r_s2 = pdist(np.vstack([X[s1[0],:], X[s2[0],:]]), metric)[0]
    #     d_s1r_s2r = pdist(np.vstack([X[s1[0],:], X[s2[-1],:]]), metric)[0]
    #     d_s2_s1 = pdist(np.vstack([X[s2[-1],:], X[s1[0],:]]), metric)[0]
    #     d_s2_s1r = pdist(np.vstack([X[s2[-1],:], X[s1[-1],:]]), metric)[0]
    #     d_s2r_s1 = pdist(np.vstack([X[s2[0],:], X[s1[0],:]]), metric)[0]
    #     d_s2r_s1r = pdist(np.vstack([X[s2[0],:], X[s1[-1],:]]), metric)[0]

    #     # if we are anchoring the first element to the start of the list then we invalidate orderings
    #     # that would move that element into a different position
    #     if anchor_first:
    #         max_val = max(d_s1_s2, d_s1_s2r, d_s1r_s2, d_s1r_s2r, d_s2_s1, d_s2_s1r, d_s2r_s1, d_s2r_s1r) + 1
    #         if s1[0] == 0:
    #             d_s1r_s2 = max_val
    #             d_s1r_s2r = max_val
    #             d_s2_s1 = max_val
    #             d_s2_s1r = max_val
    #             d_s2r_s1 = max_val
    #             d_s2r_s1r = max_val
    #         elif s2[0] == 0:
    #             d_s1_s2 = max_val
    #             d_s1_s2r = max_val
    #             d_s1r_s2 = max_val
    #             d_s1r_s2r = max_val
    #             d_s2r_s1 = max_val
    #             d_s2r_s1r = max_val

    #     # concatenete the lists in the way the minimizes the difference between
    #     # the samples at the junction
    #     best = min(d_s1_s2, d_s1_s2r, d_s1r_s2, d_s1r_s2r, d_s2_s1, d_s2_s1r, d_s2r_s1, d_s2r_s1r)
    #     if best == d_s1_s2:
    #         sets.append(s1 + s2)
    #     elif best == d_s1_s2r:
    #         sets.append(s1 + list(reversed(s2)))
    #     elif best == d_s1r_s2:
    #         sets.append(list(reversed(s1)) + s2)
    #     elif best == d_s1r_s2r:
    #         sets.append(list(reversed(s1)) + list(reversed(s2)))
    #     elif best == d_s2_s1:
    #         sets.append(s2 + s1)
    #     elif best == d_s2_s1r:
    #         sets.append(s2 + list(reversed(s1)))
    #     elif best == d_s2r_s1:
    #         sets.append(list(reversed(s2)) + s1)
    #     elif best == d_s2r_s1r:
    #         sets.append(list(reversed(s2)) + list(reversed(s1)))
    
    # return sets[-1]

# we can build better clsutering trees by doing grouping based on reducing xgb losses
def xgboost_r2(X, y, random_state=0):
    import xgboost
    X_train,X_test,y_train,y_test = sklearn.model_selection.train_test_split(X, y, random_state=random_state)

    model = xgboost.XGBRegressor(n_estimators=500)
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=10, verbose=False)

    # return the R^2 value
    return 1 - np.mean((y - model.predict(X))**2) / np.mean((y - np.mean(y))**2)

def retrain_hclustering(X, y, model=xgboost_r2, random_state=0):
    Xtmp = X.values
    dist = np.zeros((Xtmp.shape[1], Xtmp.shape[1]))
    for i in range(Xtmp.shape[1]):
        for j in range(i, Xtmp.shape[1]):
            dist[i,j] = dist[j,i] = xgboost_r2(Xtmp[:,[i,j]], y, random_state=random_state)
            
    #dist2 = np.zeros((Xtmp.shape[1], Xtmp.shape[1]))
    out = []
    for i in range(Xtmp.shape[1]):
        for j in range(i+1, Xtmp.shape[1]):
            if i != j:
                out.append(1 - (dist[i,i] + dist[j,j] - dist[i,j]) / dist[i,j])
    
    return np.array(out)

def hclust(X, y=None, linkage="complete", metric="auto", random_state=0):
    if metric == "auto":
        if y is not None:
            metric = "retrain_xgboost_r2"
    
    # build the distance matrix
    if metric == "retrain_xgboost_r2":
        dist = retrain_hclustering(X, y, model=xgboost_r2, random_state=random_state)
    else:
        if y is not None:
            warnings.warn("Ignoring the y argument passed to shap.utils.hclust since the given clustering metric is not based on label fitting!")
        if safe_isinstance(X, "pandas.core.frame.DataFrame"):
            bg_no_nan = X.values.copy()
        else:
            bg_no_nan = X.copy()
        for i in range(bg_no_nan.shape[1]):
            np.nan_to_num(bg_no_nan[:,i], nan=np.nanmean(bg_no_nan[:,i]), copy=False)
        dist = sp.spatial.distance.pdist(bg_no_nan.T + np.random.randn(*bg_no_nan.T.shape)*1e-8, metric=metric)
    # else:
    #     raise Exception("Unknown metric: " + str(metric))

    # build linkage
    if linkage == "complete":
        return sp.cluster.hierarchy.complete(dist)
    else:
        raise Exception("Unknown linkage: " + str(linkage))


from ..utils import MaskedModel
import pandas as pd
import scipy as sp
import numpy as np
import warnings
import time
from tqdm.auto import tqdm
import queue
from ..utils import assert_import, record_import_error, safe_isinstance, make_masks
from .._explanation import Explanation
from .. import maskers
from ._explainer import Explainer
from .. import links


# .shape[0] messes up pylint a lot here
# pylint: disable=unsubscriptable-object


class Partition(Explainer):
    
    def __init__(self, model, masker, *, partition_tree=None, output_names=None, link=links.identity):
        """ Uses the Partition SHAP method to explain the output of any function.

        Partition SHAP computes Shapley values recursively through a hierarchy of features, this
        hierarchy defines feature coalitions and results in the Owen values from game theory. The
        PartitionExplainer has two particularly nice properties: 1) PartitionExplainer is
        model-agnostic but when using a balanced partition tree only has quadradic exact runtime 
        (in term of the number of input features). This is in contrast to the exponential exact
        runtime of KernalExplainer or SamplingExplainer. 2) PartitionExplainer always assigns to groups of
        correlated features the credit that set of features would have had if treated as a group. This
        means if the hierarchical clustering given to PartitionExplainer groups correlated features
        together, then feature correlations are "accounted for" ... in the sense that the total credit assigned
        to a group of tightly dependent features does net depend on how they behave if their correlation
        structure was broken during the explanation's perterbation process. Note that for linear models
        the Owen values that PartitionExplainer returns are the same as the standard non-hierarchical
        Shapley values.


        Parameters
        ----------
        model : function
            User supplied function that takes a matrix of samples (# samples x # features) and
            computes the output of the model for those samples.

        masker : function or numpy.array or pandas.DataFrame or tokenizer
            The function used to "mask" out hidden features of the form `masker(mask, x)`. It takes a
            single input sample and a binary mask and returns a matrix of masked samples. These
            masked samples will then be evaluated using the model function and the outputs averaged.
            As a shortcut for the standard masking using by SHAP you can pass a background data matrix
            instead of a function and that matrix will be used for masking. Domain specific masking
            functions are available in shap such as shap.maksers.Image for images and shap.maskers.Text
            for text.

        partition_tree : None or function or numpy.array
            A hierarchical clustering of the input features represented by a matrix that follows the format
            used by scipy.cluster.hierarchy (see the notebooks/partition_explainer directory an example).
            If this is a function then the function produces a clustering matrix when given a single input
            example. If you are using a standard SHAP masker object then you can pass masker.clustering
            to use that masker's built-in clustering of the features, or if partition_tree is None then
            masker.clustering will be used by default.
        """

        super(Partition, self).__init__(model, masker, algorithm="partition")

        warnings.warn("explainers.Partition is still in an alpha state, so use with caution...")
        
        # convert dataframes
        # if safe_isinstance(masker, "pandas.core.frame.DataFrame"):
        #     masker = TabularMasker(masker)
        # elif safe_isinstance(masker, "numpy.ndarray") and len(masker.shape) == 2:
        #     masker = TabularMasker(masker)
        # elif safe_isinstance(masker, "transformers.PreTrainedTokenizer"):
        #     masker = TextMasker(masker)
        # self.masker = masker

        # TODO: maybe? if we have a tabular masker then we build a PermutationExplainer that we
        # will use for sampling
        self.input_shape = masker.shape[1:] if hasattr(masker, "shape") and not callable(masker.shape) else None
        self.output_names = output_names

        self.model = lambda x: np.array(model(x))
        self.expected_value = None
        if getattr(self.masker, "clustering", None) is None:
            raise ValueError("The passed masker must have a .clustering attribute defined! Try shap.maskers.Partition(data) for example.")
        # if partition_tree is None:
        #     if not hasattr(masker, "partition_tree"):
        #         raise ValueError("The passed masker does not have masker.clustering, so the partition_tree must be passed!")
        #     self.partition_tree = masker.clustering
        # else:
        #     self.partition_tree = partition_tree
        
        # handle higher dimensional tensor inputs
        if self.input_shape is not None and len(self.input_shape) > 1:
            self._reshaped_model = lambda x: self.model(x.reshape(x.shape[0], *self.input_shape))
        else:
            self._reshaped_model = self.model

        # if we don't have a dynamic clustering algorithm then can precowe mpute
        # a lot of information
        if not callable(self.masker.clustering):
            self._clustering = self.masker.clustering
            self._mask_matrix = make_masks(self._clustering)

    def __call__d(self, X, y=None, max_evals=1000, output_indexes=None,
                 batch_size=10, hierarchical=None, silent=False):
        """ Explain the model's prediction on the given set of instances.

        Parameters
        ----------
        X : numpy.array or pandas.DataFrame or [TODO] pytorch or TF array
            An array of samples for which to compute explanations.

        y : numpy.array or pandas.DataFrame or [TODO] pytorch or TF array
            An optional parallel array to X that provides the labels of the samples.

        npartitions : int
            How many times we partition the data, and hence evaluate the underlying model. More partitions lead to more
            fine-grained explanations, but require more run-time.

        hierarchical : None (default) or bool
            Whether to leave all interaction effects on the internal nodes of the partition tree, or to instead
            use Shapley values to split them up among lower nodes. By default this behavior is chosen
            automatically by the masker object type, but this can be overridden by setting to True of False directly.
        """

        npartitions = max_evals // 2
        
        # convert dataframes
        feature_names = None
        if safe_isinstance(X, "pandas.core.frame.DataFrame"):
            feature_names = list(X.columns)
            X = X.values
        elif safe_isinstance(X, "pandas.core.series.Series"):
            X = X.values

        # get hierarchical setting
        if hierarchical is None:
            hierarchical = getattr(self.masker, "hierarchical", False)
            
        # convert strings
        if type(X[0]) is str:
            #X_str = X
            tokenized = [self.masker.tokenize(s) for s in X]
            X = np.array([t["input_ids"] for t in tokenized])
            #offsets = [t["offset_mapping"] for t in tokenized]
        
        # handle list inputs that are not multiple inputs by auto-wrapping them as np.arrays
        if type(X) is list and not hasattr(X[0], "__len__"):
            X = np.array(X)
        elif type(X) is np.array and type(X[0]) is list:
            X = np.array([np.array(v) for v in X])
        
        # check we got just a single instance (TODO: use the masker to make this check robust to multi-dim inputs)
        # single_instance = False
        # if not hasattr(X[0], "__len__"):
        #     single_instance = True
        #     X = X.reshape(1,-1)
            
        # handle higher dimensional tensor inputs
        orig_shape = X.shape[1:]
        if len(X.shape) > 2:
            X = X.reshape(X.shape[0], -1)
            self._reshaped_model = lambda X: self.model(X.reshape(X.shape[0], *orig_shape))
        else:
            self._reshaped_model = self.model
            
        out = []
        out_inds = []
        pbar = None
        start_time = time.time()
        for i in range(X.shape[0]):
            row_out,oinds = self.explain(X[i], hierarchical, npartitions, output_indexes, 0, batch_size, silent)
            out.append(row_out)
            if output_indexes is not None:
                out_inds.append(oinds)
            if pbar is None and time.time() - start_time > 5:
                pbar = tqdm(total=X.shape[0], disable=silent, leave=False)
                pbar.update(i)
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()
        out = np.array(out)
        out_inds = np.array(out_inds)
        out_len = 0 if self.output_shape == tuple() else self.output_shape[0]
        if not hierarchical:
            if out_len == 0:
                out = out.reshape(out.shape[0], *orig_shape)
            else:
                out = out.reshape(out.shape[0], *orig_shape, out_len)
        
        # the output shape
#         out_len = 0 if self.output_shape == tulpe() else out_line
#         if self.multi_output:
#             out_len = len(self.curr_expected_value) if output_indexes is None else output_indexes_len(output_indexes)
#         output_shape = tuple(range(out_len))
        
        if output_indexes is not None:
            output_names = [[self.output_names[j] for j in out_inds[i]] for i in range(out_len)]
        else:
            output_names = self.output_names
        
        return Explanation(
            out, self.expected_value, X, input_names=feature_names, output_shape=self.output_shape,
            output_indexes=out_inds, output_names=output_names
        )
        
#         if self.multi_output:
#             n = np.prod(orig_shape)
#             out_len = len(self.curr_expected_value) if output_indexes is None else output_indexes_len(output_indexes)
#             out = out.reshape(out.shape[0], np.prod(orig_shape), out_len)
#             out = [out[:,:,i].reshape(out.shape[0], *orig_shape) for i in range(out_len)]
                
#             if output_indexes is not None:
#                 return out, out_inds
#             else:
#                 return out
#         else:
#             out = out.reshape(out.shape[0], *orig_shape)
#             if single_instance:
#                 return out[0]
#             else:
#                 return out
            
            
            
            
            
            
            
            
            
#         # convert strings
#         if type(X[0]) is str:
#             X_str = X
#             tokenized = [self.masker.tokenize(s) for s in X]
#             X = np.array([t["input_ids"] for t in tokenized])
#             offsets = [t["offset_mapping"] for t in tokenized]
            
#         if hierarchical:
#             v = self.hierarchical_shap_values(
#                 X, npartitions=npartitions, interaction_tolerance=interaction_tolerance,
#                 batch_size=batch_size, output_indexes=output_indexes
#             )
#             out_inds = None
#             if output_indexes is not None:
#                 out_inds = v[1]
#                 v = v[0]
#             output_shape = tuple()
#             if type(v) is list:
#                 output_shape = (len(v),)
#                 v = np.stack(v, axis=-1) # put outputs at the end
#             output_names = None
#             if out_inds is not None:
#                 output_names = [[self.output_names[j] for j in out_inds[i]] for i in range(v.shape[0])]
#             e = Explanation(
#                 self.expected_value, v, X, input_names=feature_names, output_shape=output_shape,
#                 output_indexes=out_inds, output_names=output_names
#             )
#         else:
#             v = self.shap_values(
#                 X, npartitions=npartitions, interaction_tolerance=interaction_tolerance,
#                 batch_size=batch_size, output_indexes=output_indexes
#             )
#             out_inds = None
#             if output_indexes is not None:
#                 out_inds = v[1]
#                 v = v[0]
#             output_shape = tuple()
#             if type(v) is list:
#                 output_shape = (len(v),)
#                 v = np.stack(v, axis=-1) # put outputs at the end
#             output_names = None
#             if out_inds is not None:
#                 output_names = [[self.output_names[j] for j in out_inds[i]] for i in range(v.shape[0])]
#             e = Explanation(
#                 self.expected_value, v, X, input_names=feature_names, output_shape=output_shape,
#                 output_indexes=out_inds, output_names=output_names
#             )
# #         else:
# #             assert False, "hierarchical not implemented yet for __call__... simple fix TBD."
#         return e

    # def create_cluster_matrices(self, partition_tree):
    #     """ Build clustering dependent reuseable variables.
    #     """
        
    #     self._mask_matrix = make_masks(partition_tree)
    #     self.merge_clusters = -np.ones((2 * partition_tree.shape[0] + 1, 3), dtype=np.int64)
    #     self.merge_clusters_score = np.zeros(self.merge_clusters.shape[0])
    #     self.merge_clusters[partition_tree.shape[0] + 1:,:2] = partition_tree[:,:2].astype(np.int64)
    #     self.merge_clusters_score[partition_tree.shape[0] + 1:] = partition_tree[:,2]
    #     for i in range(self.merge_clusters.shape[0]):
    #         if self.merge_clusters[i,0] < 0:
    #             self.merge_clusters[i,2] = 1
    #         else:
    #             self.merge_clusters[i,2] = self.merge_clusters[self.merge_clusters[i,0],2] + self.merge_clusters[self.merge_clusters[i,1],2]
    
    def hierarchical_shap_values(self, x, npartitions=100, output_indexes=None,
                                 context=0, batch_size=10, silent=False):
        
        # convert dataframes
        if safe_isinstance(x, "pandas.core.series.Series"):
            x = x.values
        elif safe_isinstance(x, "pandas.core.frame.DataFrame"):
            x = x.values
            
        # convert strings
        if type(x[0]) is str:
            x_str = x
            tokenized = [self.masker.tokenize(s) for s in x]
            x = np.array([np.array(t["input_ids"]) for t in tokenized])
            #offsets = [t["offset_mapping"] for t in tokenized]

        # handle list inputs that are not multiple inputs by auto-wrapping them as np.arrays
        if type(x) is list and not hasattr(x[0], "__len__"):
            x = np.array(x)

        single_instance = False
        if not hasattr(x[0], "__len__"):
            single_instance = True
            x = x.reshape(1, -1)
        
        # handle higher dimensional tensor inputs
        if hasattr(x[0][0], "__len__"):
            orig_shape = x[0].shape
            x = x.reshape(len(x), -1)
            self._reshaped_model = lambda x: self.model(x.reshape(len(x), *orig_shape))
        else:
            self._reshaped_model = self.model
            orig_shape = None

        out = []
        out_inds = []
        pbar = None
        start_time = time.time()
        for i in range(x.shape[0]):
            row_out,oinds = self.explain(
                x[i], "hierarchical", npartitions, output_indexes, -1, context, batch_size, silent
            )
            out.append(row_out)
            if output_indexes is not None:
                out_inds.append(oinds)
            if pbar is None and time.time() - start_time > 5:
                pbar = tqdm(total=x.shape[0], disable=silent, leave=False)
                pbar.update(i)
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()
        out = np.array(out)
        out_inds = np.array(out_inds)

        if single_instance:
            return out[0]
        else:
            return out
        
    def shap_values(self, x, npartitions=100, interaction_tolerance=0, output_indexes=None,
                    batch_size=10, silent=False):
        
        # convert dataframes
        if safe_isinstance(x, "pandas.core.series.Series"):
            x = x.values
        elif safe_isinstance(x, "pandas.core.frame.DataFrame"):
            x = x.values
        
        # handle list inputs that are not multiple inputs by auto-wrapping them as np.arrays
        if type(x) is list and not hasattr(x[0], "__len__"):
            x = np.array(x)
        elif type(x) is np.array and type(x[0]) is list:
            x = np.array([np.array(v) for v in x])
        
        single_instance = False
        if not hasattr(x[0], "__len__"):
            single_instance = True
            x = x.reshape(1,-1)
            
        # handle higher dimensional tensor inputs
        orig_shape = x.shape[1:]
        if len(x.shape) > 2:
            x = x.reshape(x.shape[0], -1)
            self._reshaped_model = lambda x: self.model(x.reshape(x.shape[0], *orig_shape))
        else:
            self._reshaped_model = self.model
            
        out = []
        out_inds = []
        pbar = None
        start_time = time.time()
        for i in range(x.shape[0]):
            row_out,oinds = self.explain(
                x[i], "marginal", npartitions, output_indexes,
                interaction_tolerance, 0, batch_size, silent
            )
            out.append(row_out)
            if output_indexes is not None:
                out_inds.append(oinds)
            if pbar is None and time.time() - start_time > 5:
                pbar = tqdm(total=x.shape[0], disable=silent, leave=False)
                pbar.update(i)
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()
        out = np.array(out)
        out_inds = np.array(out_inds)
        
        if self.multi_output:
            n = np.prod(orig_shape)
            out_len = len(self.curr_expected_value) if output_indexes is None else output_indexes_len(output_indexes)
            out = out.reshape((out.shape[0], np.prod(orig_shape), out_len))
            out = [out[:,:,i].reshape(out.shape[0], *orig_shape) for i in range(out_len)]
                
            if output_indexes is not None:
                return out, out_inds
            else:
                return out
        else:
            out = out.reshape(out.shape[0], *orig_shape)
            if single_instance:
                return out[0]
            else:
                return out

    def explain_row(self, *row_args, max_evals, main_effects, error_bounds, batch_size, silent):
        """ Explains a single row and returns the tuple (row_values, row_expected_values, row_mask_shapes).
        """

        # build a masked version of the model for the current input sample
        fm = MaskedModel(self.model, self.masker, self.link, *row_args)

        if max_evals == "auto":
            max_evals = 100

        if callable(self.masker.clustering):
            self._clustering = self.masker.clustering(*row_args)
            self._mask_matrix = make_masks(self._clustering)
        self.multi_output = False
        # allocate space for our outputs
        if self.multi_output:
            if output_indexes is not None:
                out_shape = (2*self._clustering.shape[0]+1, output_indexes_len(output_indexes))
            else:
                out_shape = (2*self._clustering.shape[0]+1, len(self.curr_expected_value))
        else:
            out_shape = (2*self._clustering.shape[0]+1,)
        self.values = np.zeros(out_shape)
        self.dvalues = np.zeros(out_shape)

        output_indexes = None
        fixed_context = 1
        output_indexes, base_value = self.owen(fm, max_evals // 2, output_indexes, fixed_context, batch_size, silent)

        if False:
            if self.multi_output:
                return [self.dvalues[:,i] for i in range(self.dvalues.shape[1])], oinds
            else:
                return self.dvalues.copy(), oinds   
        else:
            # drop the interaction terms down onto self.values
            self.values[:] = self.dvalues
            M = len(fm)
            def lower_credit(i, value=0):
                if i < M:
                    self.values[i] += value
                    return
                li = int(self._clustering[i-M,0])
                ri = int(self._clustering[i-M,1])
                group_size = int(self._clustering[i-M,3])
                lsize = int(self._clustering[li-M,3]) if li >= M else 1
                rsize = int(self._clustering[ri-M,3]) if ri >= M else 1
                assert lsize+rsize == group_size
                self.values[i] += value
                lower_credit(li, self.values[i] * lsize / group_size)
                lower_credit(ri, self.values[i] * rsize / group_size)
            lower_credit(len(self.dvalues) - 1)
            

        return {
            "values": self.values[:len(fm)].copy(),
            "expected_values": base_value,
            "mask_shapes": fm.mask_shapes,
            "main_effects": None,
            "hierarchical_values": self.dvalues.copy(),
            "clustering": self._clustering
        }

    def owen(self, fm, npartitions, output_indexes, fixed_context, batch_size, silent):
        """ Compute a nested set of recursive Owen values based on an ordering recursion.
        """
        
        #f = self._reshaped_model
        #r = self.masker
        #masks = np.zeros(2*len(inds)+1, dtype=np.int)
        M = len(fm)
        m00 = np.zeros(M, dtype=np.bool)
        f00 = fm(m00.reshape(1,-1))[0]
        base_value = f00
        f11 = fm(~m00.reshape(1,-1))[0]
        #f11 = self._reshaped_model(r(~m00, x)).mean(0)
        ind = len(self.dvalues)-1

        # make sure output_indexes is a list of indexes
        if output_indexes is not None:
            assert self.multi_output, "output_indexes is only valid for multi-output models!"
            inds = output_indexes.apply(f11, 0)
            out_len = output_indexes_len(output_indexes)
            if output_indexes.startswith("max("):
                output_indexes = np.argsort(-f11)[:out_len]
            elif output_indexes.startswith("min("):
                output_indexes = np.argsort(f11)[:out_len]
            elif output_indexes.startswith("max(abs("):
                output_indexes = np.argsort(np.abs(f11))[:out_len]
        
            f00 = f00[output_indexes]
            f11 = f11[output_indexes]
        
        q = queue.PriorityQueue()
        q.put((0, 0, (m00, f00, f11, ind, 1.0)))
        eval_count = 0
        total_evals = min(npartitions, (M-1)*M) # TODO: (len(x)-1)*len(x) is only right for balanced partition trees
        pbar = None
        start_time = time.time()
        while not q.empty():

            # if we passed our execution limit then leave everything else on the internal nodes
            if eval_count >= npartitions:
                while not q.empty():
                    m00, f00, f11, ind, weight = q.get()[2]
                    self.dvalues[ind] += (f11 - f00) * weight
                break

            # create a batch of work to do
            batch_args = []
            batch_masks = []
            while not q.empty() and len(batch_masks) < batch_size and eval_count < npartitions:
                
                # get our next set of arguments
                m00, f00, f11, ind, weight = q.get()[2]

                # get the left and right children of this cluster
                lind = int(self._clustering[ind-M, 0]) if ind >= M else -1
                rind = int(self._clustering[ind-M, 1]) if ind >= M else -1
                distance = self._clustering[ind-M, 2] if ind >= M else -1

                # check if we are a leaf node (or other negative distance cluster) and so should terminate our decent
                if distance < 0:
                    self.dvalues[ind] += (f11 - f00) * weight
                    continue

                # build the masks
                m10 = m00.copy() # we separate the copy from the add so as to not get converted to a matrix
                m10[:] += self._mask_matrix[lind, :]
                m01 = m00.copy()
                m01[:] += self._mask_matrix[rind, :]
                
                batch_args.append((m00, m10, m01, f00, f11, ind, lind, rind, weight))
                batch_masks.append(m10)
                batch_masks.append(m01)
            
            batch_masks = np.array(batch_masks)
                
            # run the batch
            if len(batch_args) > 0:
                fout = fm(batch_masks)
                if output_indexes is not None:
                    fout = fout[:,output_indexes]
                    
                eval_count += len(batch_masks)
                    
                if pbar is None and time.time() - start_time > 5:
                    pbar = tqdm(total=total_evals, disable=silent, leave=False)
                    pbar.update(eval_count)
                if pbar is not None:
                    pbar.update(len(batch_masks))
            
            # use the results of the batch to add new nodes
            for i in range(len(batch_args)):
                
                m00, m10, m01, f00, f11, ind, lind, rind, weight = batch_args[i]
                
                # get the evaluated model output on the two new masked inputs
                f10 = fout[2*i]
                f01 = fout[2*i+1]

                new_weight = weight
                if fixed_context is None:
                    new_weight /= 2
                elif fixed_context == 0:
                    self.dvalues[ind] += (f11 - f10 - f01 + f00) * weight # leave the interaction effect on the internal node
                elif fixed_context == 1:
                    self.dvalues[ind] -= (f11 - f10 - f01 + f00) * weight # leave the interaction effect on the internal node

                if fixed_context is None or fixed_context == 0:
                    # recurse on the left node with zero context
                    args = (m00, f00, f10, lind, new_weight)
                    q.put((-np.max(np.abs(f10 - f00)) * new_weight, np.random.randn(), args))

                    # recurse on the right node with zero context
                    args = (m00, f00, f01, rind, new_weight)
                    q.put((-np.max(np.abs(f01 - f00)) * new_weight, np.random.randn(), args))

                if fixed_context is None or fixed_context == 1:
                    # recurse on the left node with one context
                    args = (m01, f01, f11, lind, new_weight)
                    q.put((-np.max(np.abs(f11 - f01)) * new_weight, np.random.randn(), args))

                    # recurse on the right node with one context
                    args = (m10, f10, f11, rind, new_weight)
                    q.put((-np.max(np.abs(f11 - f10)) * new_weight, np.random.randn(), args))
                    
        if pbar is not None:
            pbar.close()
        
        return output_indexes, base_value
        

    def explain(self, x, hierarchical, npartitions, output_indexes, context, batch_size, silent):
        
        if getattr(self.masker, "variable_background", False) or getattr(self, "curr_expected_value", None) is None:
            self.eval_time = time.time()
            self.curr_expected_value = self._reshaped_model(self.masker(np.zeros(x.shape, dtype=np.bool), x)).mean(0)
            self.eval_time = time.time() - self.eval_time
            if hasattr(self.curr_expected_value, 'shape') and len(self.curr_expected_value.shape) > 0:
                self.multi_output = True
                out_len = len(self.curr_expected_value) if output_indexes is None else output_indexes_len(output_indexes)
                self.output_shape = (out_len,)
            else:
                self.multi_output = False
                self.output_shape = tuple()
            
            if not getattr(self.masker, "variable_background", False):
                self.expected_value = self.curr_expected_value
            
        
        if callable(self.masker.clustering):
            self.create_cluster_matrices(self.masker.clustering(x))
        
        # allocate space for our outputs
        if self.multi_output:
            if output_indexes is not None:
                out_shape = (self.merge_clusters.shape[0], output_indexes_len(output_indexes))
            else:
                out_shape = (self.merge_clusters.shape[0], len(self.curr_expected_value))
        else:
            out_shape = (self.merge_clusters.shape[0],)
        self.values = np.zeros(out_shape)
        self.dvalues = np.zeros(out_shape)
        oinds = self.owen(x, npartitions, output_indexes, hierarchical, context, batch_size, silent)

        if hierarchical:
            if self.multi_output:
                return [self.dvalues[:,i] for i in range(self.dvalues.shape[1])], oinds
            else:
                return self.dvalues.copy(), oinds   
        else:
            # drop the interaction terms down onto self.values
            self.values[:] = self.dvalues
            M = len(x)
            def lower_credit(i, value=0):
                if i < M:
                    self.values[i] += value
                    return
                li = self.merge_clusters[i,0]
                ri = self.merge_clusters[i,1]
                group_size = self.merge_clusters[i,2]
                lsize = self.merge_clusters[li,2]
                rsize = self.merge_clusters[ri,2]
                assert lsize+rsize == group_size
                self.values[i] += value
                lower_credit(li, self.values[i] * lsize / group_size)
                lower_credit(ri, self.values[i] * rsize / group_size)
            lower_credit(len(self.dvalues) - 1)
            
            return self.values[:len(x)].copy(), oinds
        
    def single_context_owen(self, x, npartitions, output_indexes, interaction_tolerance, context, silent):
        """ Compute a nested set of recursive Owen values.
        """
        
        f = self._reshaped_model
        r = self.masker
        m00 = np.zeros(self._mask_matrix.shape[1], dtype=np.bool)
        f00 = self.curr_expected_value
        f11 = self._reshaped_model(r(~m00, x)).mean(0)
        ind = len(self.values)-1

        if context == 1:
            tmp = f00
            f00 = f11
            f11 = tmp

        q = queue.PriorityQueue()
        q.put((0, (m00, f00, f11, ind)))
        eval_count = 0
        total_evals = min(npartitions, (len(x)-1)*len(x))
        pbar = None
        start_time = time.time()
        while not q.empty():

            # if we passed our execution limit then leave everything else on the internal nodes
            if eval_count > npartitions:
                while not q.empty():
                    m00, f00, f11, ind = q.get()[1]
                    self.dvalues[ind] = f11 - f00
                break

            # get our next set of arguments
            m00, f00, f11, ind = q.get()[1]

            # get the left are right children of this cluster
            lind = self.merge_clusters[ind, 0]
            rind = self.merge_clusters[ind, 1]

            # check if we are a leaf node or terminated our decent early and dumping credit at an internal node
            if lind < 0:
                self.dvalues[ind] = f11 - f00
                continue

            # build the masks
            m10 = m00.copy() # we separate the copy from the add so as to not get converted to a matrix
            m10[:] += self._mask_matrix[lind, :]
            m01 = m00.copy()
            m01[:] += self._mask_matrix[rind, :]

            # evaluate the model on the two new masked inputs
            if context == 0:
                # assert r(m10, x)[0][0] == 101
                # assert r(m01, x)[0][0] == 101
                f10 = f(r(m10, x)).mean(0)
                f01 = f(r(m01, x)).mean(0)
            else:
                f10 = f(r(~m10, x)).mean(0)
                f01 = f(r(~m01, x)).mean(0)
            
            # update our progress indicator
            eval_count += 2
            if pbar is None and time.time() - start_time > 5:
                pbar = tqdm(total=total_evals, disable=silent, leave=False)
                pbar.update(eval_count)
            if pbar is not None:
                pbar.update(2)

            # update our dividends
            self.dvalues[ind] += f11 - f01 - f10 + f00

            # recurse on the left node
            args = (m00, f00, f10, lind)
            q.put((-abs(f10 - f00), args)) #/ np.sqrt(m10.sum()

            # recurse on the right node
            args = (m00, f00, f01, rind)
            q.put((-abs(f01 - f00), args))
        
        if pbar is not None:
            pbar.close()

        if context == 1:
            self.dvalues *= -1

        return None
    
    def owen4(self, x, npartitions, output_indexes, interaction_tolerance, context, batch_size, silent):
        """ Compute a nested set of recursive Owen values based on an ordered recursion of paried permutations.

        TODO: This is not done yet, but when it is done it should outperform our current solution of just using
        the PermutationExplainer for tabular data.
        """
        
        f = self._reshaped_model
        r = self.masker
        m00 = np.zeros(self._mask_matrix.shape[1], dtype=np.bool)
        f00 = self.curr_expected_value
        f11 = self._reshaped_model(r(~m00, x)).mean(0)
        ind = len(self.values)-1
        scaling = np.ones(4*npartitions)
        sparent = np.ones(4*npartitions, dtype=np.int) * -1
        scount = 0
        required_evals = 0

        # make sure output_indexes is a list of indexes
        if output_indexes is not None:
            assert self.multi_output, "output_indexes is only valid for multi-output models!"
            
            out_len = output_indexes_len(output_indexes)
            if output_indexes.startswith("max("):
                output_indexes = np.argsort(-f11)[:out_len]
            elif output_indexes.startswith("min("):
                output_indexes = np.argsort(f11)[:out_len]
            elif output_indexes.startswith("max(abs("):
                output_indexes = np.argsort(np.abs(f11))[:out_len]
        
            f00 = f00[output_indexes]
            f11 = f11[output_indexes]
        
        q = queue.PriorityQueue()
        q.put((0, 0, 0, (m00, f00, f11, ind, scount, -1, True, self.merge_clusters[-1, 2], 0), 1.0))
        
        scount += 1
        eval_count = 0
        total_evals = min(npartitions, (len(x)-1)*len(x)) # TODO: (len(x)-1)*len(x) is only right for balanced partition trees
        pbar = None
        start_time = time.time()
        done_list = []
        skip_next_reverse = False
        while not q.empty():

            # if we passed our execution limit then leave everything else on the internal nodes
            # if eval_count >= npartitions:
            #     while not q.empty():
            #         done_list.append(q.get())
            #         # m00, f00, f11, ind, sind, twin_ind = q.get()[2]
            #         # self.dvalues[ind] += (f11 - f00) * compute_weight(sind, sparent, scaling)
            #     break

            # create a batch of work to do
            batch_args = []
            batch_data = []
            batch_positions = []
            batch_pos = 0
            will_eval = 0
            while not q.empty() and len(batch_data) < batch_size:# and eval_count <= npartitions:
                
                # get our next set of arguments
                parts = q.get()
                m00, f00, f11, ind, sind, twin_ind, forward, eval_cost, depth = parts[3]
                prev_weight = parts[4]
                curr_weight = compute_weight(sind, sparent, scaling)

                # if our weight has changed since we last added this node then we put the node back on the
                # queue with an updated weight and then start over
                if abs(prev_weight - curr_weight) > 1e-5:
                    q.put((parts[0], parts[1] * curr_weight / prev_weight, parts[2], parts[3], curr_weight))
                    continue

                # get the left are right children of this cluster
                lind = self.merge_clusters[ind, 0]
                rind = self.merge_clusters[ind, 1]

                # if we shouldn't be expanding reverse nodes anymore then we skip all of them
                if parts[0] == 1 and skip_next_reverse:
                    skip_next_reverse = False
                    continue
                
                if eval_cost > 0:
                    if npartitions - required_evals < eval_cost:
                        skip_next_reverse = True # we skip the next reverse node since it is coupled with us
                        continue
                    required_evals += eval_cost
                
                

                # if we have a twin, then we lower that twin's weight since it now needs to share it with us :)
                
                if twin_ind >= 0:
                    #print(npartitions, required_evals, eval_cost, ind, forward)
                    # cluster_size = self.merge_clusters[ind, 2]
                    # evals_left = npartitions - required_evals
                    # print("cs", cluster_size, evals_left, ind, self.merge_clusters.shape[0])
                    # if evals_left < cluster_size or not (ind == self.merge_clusters[-1,0] or ind == self.merge_clusters[-1,1]):
                    #     continue # we don't expand reverse orderings we can't finish...
                    # else:
                    #     print("ASDFASDF", required_evals, cluster_size)
                    scaling[twin_ind] /= 2

                # print()
                # print("run", ind, "forward" if forward else "reverse", parts[0], twin_ind, lind, rind, self.merge_clusters[-1,:])
                # print(m00 * 1)
                # print(m00 + self._mask_matrix[ind, :].toarray()[0] * 1)
                # print(npartitions, required_evals, eval_cost)

                
                
                
                # check if we are a leaf node and so terminating our decent
                if lind < 0:
                    done_list.append(parts)
                    continue

                # build the masks
                m10 = m00.copy() # we separate the copy from the add so as to not get converted to a matrix
                m10[:] += self._mask_matrix[lind, :]
                m01 = m00.copy()
                m01[:] += self._mask_matrix[rind, :]
                
                # print(m10 * 1, twin_ind)
                # print(m01 * 1, twin_ind)

                
                batch_args.append((m00, m10, m01, f00, f11, ind, lind, rind, sind, twin_ind, forward, eval_cost, depth))
                
                d = r(m10, x)
                batch_data.append(d)
                batch_positions.append(batch_pos)
                batch_pos += d.shape[0]
                will_eval += 1
                
                d = r(m01, x)
                batch_data.append(d)
                batch_positions.append(batch_pos)
                batch_pos += d.shape[0]
                will_eval += 1
                
            batch_positions.append(batch_pos)
                
            # run the batch
            if len(batch_args) > 0:
                if safe_isinstance(batch_data[0], "pandas.core.frame.DataFrame"):
                    fout = f(pd.concat(batch_data, axis=0))
                    #print(pd.concat(batch_data, axis=0))
                else:
                    fout = f(np.concatenate(batch_data, axis=0))
                    #print(np.concatenate(batch_data, axis=0))
                if output_indexes is not None:
                    fout = fout[:,output_indexes]
                    
                eval_count += len(batch_data)
                    
                if pbar is None and time.time() - start_time > 5:
                    pbar = tqdm(total=total_evals, disable=silent, leave=False)
                    pbar.update(eval_count)
                if pbar is not None:
                    pbar.update(len(batch_data))
            
            # use the results of the batch to add new nodes
            for i in range(len(batch_args)):
                
                m00, m10, m01, f00, f11, ind, lind, rind, sind, twin_ind, forward, eval_cost, depth = batch_args[i]
                
                # evaluate the model on the two new masked inputs
                f10 = fout[batch_positions[2*i]:batch_positions[2*i+1]].mean(0)
                f01 = fout[batch_positions[2*i+1]:batch_positions[2*i+2]].mean(0)

                weight = compute_weight(sind, sparent, scaling)
                lsize = self.merge_clusters[lind, 2]
                rsize = self.merge_clusters[rind, 2]

                if forward:

                    # recurse on the left node with zero context
                    args = (m00, f00, f10, lind, scount, -1, True, 0, depth+1)
                    q.put((0, -np.max(np.abs(f10 - f00)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 1.0
                    scount += 1

                    # recurse on the right node with one context
                    args = (m10, f10, f11, rind, scount, -1, True, 0, depth+1)
                    q.put((0, -np.max(np.abs(f11 - f10)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 1.0
                    scount += 1

                    # recurse on the right node with zero context
                    # this is the node that when it gets pulled off "pays" for the other nodes
                    args = (m00, f00, f01, rind, scount, scount-1, False, lsize+rsize, depth+1) # NOTE: the fourth to last arg is the scaling index of the other node that needs its weight reduced when we pull this one off
                    interaction_value = 1#f11 - f10 - f01 + f00
                    q.put((1, -np.max(np.abs(interaction_value)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 0.5
                    scount += 1

                    # recurse on the left node with one context
                    args = (m01, f01, f11, lind, scount, scount-3, False, 0, depth+1) 
                    q.put((1, -np.max(np.abs(interaction_value)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 0.5
                    scount += 1

                else:
                    
                    # recurse on the right node with zero context
                    args = (m00, f00, f01, rind, scount, -1, False, 0, depth+1)
                    q.put((0, -np.max(np.abs(f01 - f00)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 1.0
                    scount += 1

                    # recurse on the left node with one context
                    args = (m01, f01, f11, lind, scount, -1, False, 0, depth+1) # NOTE: the last arg is the scaling index of the other node that needs its weight reduced when we pull this one off
                    q.put((0, -np.max(np.abs(f10 - f00)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 1.0
                    scount += 1

                    # recurse on the left node with zero context
                    # this is the node that when it gets pulled off "pays" for the other nodes
                    args = (m00, f00, f10, lind, scount, scount - 1, True, lsize+rsize, depth+1)
                    interaction_value = 1#f11 - f10 - f01 + f00
                    q.put((1, -np.max(np.abs(interaction_value)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 0.5
                    scount += 1

                    # recurse on the right node with one context
                    args = (m10, f10, f11, rind, scount, scount - 3, True, 0, depth+1)
                    q.put((1, -np.max(np.abs(interaction_value)) / 2**(depth+1), scount, args, weight))
                    sparent[scount] = sind
                    scaling[scount] = 0.5
                    scount += 1
        
        # finish any undone twin scaling
        # for parts in done_list:
        #     twin_ind = parts[2][5]
        #     if twin_ind >= 0:
        #         scaling[twin_ind] /= 2
        
        # dump the values from the list of done nodes into dvalues
        count = 0
        for parts in done_list:
            m00, f00, f11, ind, sind, twin_ind, forward, eval_cost, depth = parts[3]
            lind = self.merge_clusters[ind, 0]
            if twin_ind == -1 or lind < 0:
                # if lind >= 0:
                #     print("internal!!", ind)
                # else:
                #     print(ind)
                # print(ind, twin_ind, f00, f11, sind)
                # print(m00 + self._mask_matrix[ind, :] + 0)
                # print("-")
                # print(m00.reshape(1,-1) + 0)
                # print()
                self.dvalues[ind] += (f11 - f00) * compute_weight(sind, sparent, scaling)
                count += 1
            else:
                pass#self.dvalues[ind] += (f11 - f00) * compute_weight(sind, sparent, scaling)
        # print("count", count)
        if pbar is not None:
            pbar.close()
        
        return output_indexes



    def owen3(self, x, npartitions, output_indexes, hierarchical, context, batch_size, silent):
        """ Compute a nested set of recursive Owen values based on an ordering recursion.
        """
        
        f = self._reshaped_model
        r = self.masker
        m00 = np.zeros(self._mask_matrix.shape[1], dtype=np.bool)
        f00 = self.curr_expected_value
        f11 = self._reshaped_model(r(~m00, x)).mean(0)
        ind = len(self.values)-1

        # make sure output_indexes is a list of indexes
        if output_indexes is not None:
            assert self.multi_output, "output_indexes is only valid for multi-output models!"
            
            out_len = output_indexes_len(output_indexes)
            if output_indexes.startswith("max("):
                output_indexes = np.argsort(-f11)[:out_len]
            elif output_indexes.startswith("min("):
                output_indexes = np.argsort(f11)[:out_len]
            elif output_indexes.startswith("max(abs("):
                output_indexes = np.argsort(np.abs(f11))[:out_len]
        
            f00 = f00[output_indexes]
            f11 = f11[output_indexes]
        
        q = queue.PriorityQueue()
        q.put((0, 0, (m00, f00, f11, ind, 1.0)))
        eval_count = 0
        total_evals = min(npartitions, (len(x)-1)*len(x)) # TODO: (len(x)-1)*len(x) is only right for balanced partition trees
        pbar = None
        start_time = time.time()
        while not q.empty():

            # if we passed our execution limit then leave everything else on the internal nodes
            if eval_count >= npartitions:
                while not q.empty():
                    m00, f00, f11, ind, weight = q.get()[2]
                    self.dvalues[ind] += (f11 - f00) * weight
                break

            # create a batch of work to do
            batch_args = []
            batch_data = []
            batch_positions = []
            batch_pos = 0
            while not q.empty() and len(batch_data) < batch_size and eval_count < npartitions:
                
                # get our next set of arguments
                m00, f00, f11, ind, weight = q.get()[2]

                # get the left are right children of this cluster
                lind = self.merge_clusters[ind, 0]
                rind = self.merge_clusters[ind, 1]

                # check if we are a leaf node and so terminating our decent
                if lind < 0:
                    self.dvalues[ind] += (f11 - f00) * weight
                    continue

                # build the masks
                m10 = m00.copy() # we separate the copy from the add so as to not get converted to a matrix
                m10[:] += self._mask_matrix[lind, :]
                m01 = m00.copy()
                m01[:] += self._mask_matrix[rind, :]
                
                batch_args.append((m00, m10, m01, f00, f11, ind, lind, rind, weight))
                
                d = r(m10, x)
                batch_data.append(d)
                batch_positions.append(batch_pos)
                batch_pos += d.shape[0]
                
                d = r(m01, x)
                batch_data.append(d)
                batch_positions.append(batch_pos)
                batch_pos += d.shape[0]
                
            batch_positions.append(batch_pos)
                
            # run the batch
            if len(batch_args) > 0:
                if safe_isinstance(batch_data[0], "pandas.core.frame.DataFrame"):
                    fout = f(pd.concat(batch_data, axis=0))
                else:
                    fout = f(np.concatenate(batch_data, axis=0))
                if output_indexes is not None:
                    fout = fout[:,output_indexes]
                    
                eval_count += len(batch_data)
                    
                if pbar is None and time.time() - start_time > 5:
                    pbar = tqdm(total=total_evals, disable=silent, leave=False)
                    pbar.update(eval_count)
                if pbar is not None:
                    pbar.update(len(batch_data))
            
            # use the results of the batch to add new nodes
            for i in range(len(batch_args)):
                
                m00, m10, m01, f00, f11, ind, lind, rind, weight = batch_args[i]
                
                # evaluate the model on the two new masked inputs
                f10 = fout[batch_positions[2*i]:batch_positions[2*i+1]].mean(0)
                f01 = fout[batch_positions[2*i+1]:batch_positions[2*i+2]].mean(0)

                new_weight = weight
                if not hierarchical:
                    new_weight /= 2
                else:
                    self.dvalues[ind] += (f11 - f10 - f01 + f00) * weight # leave the interaction effect on the internal node

                # recurse on the left node with zero context
                args = (m00, f00, f10, lind, new_weight)
                q.put((-np.max(np.abs(f10 - f00)) * new_weight, np.random.randn(), args))

                # recurse on the right node with zero context
                args = (m00, f00, f01, rind, new_weight)
                q.put((-np.max(np.abs(f01 - f00)) * new_weight, np.random.randn(), args))

                if not hierarchical:
                    # recurse on the left node with one context
                    args = (m01, f01, f11, lind, new_weight)
                    q.put((-np.max(np.abs(f11 - f01)) * new_weight, np.random.randn(), args))

                    # recurse on the right node with one context
                    args = (m10, f10, f11, rind, new_weight)
                    q.put((-np.max(np.abs(f11 - f10)) * new_weight, np.random.randn(), args))
        if pbar is not None:
            pbar.close()
        
        return output_indexes
    
def compute_weight(sind, sparent, scaling):
    weight = 1.0
    while sind >= 0:
        weight *= scaling[sind]
        sind = sparent[sind]
    return weight

def output_indexes_len(output_indexes):
    if output_indexes.startswith("max("):
        return int(output_indexes[4:-1])
    elif output_indexes.startswith("min("):
        return int(output_indexes[4:-1])
    elif output_indexes.startswith("max(abs("):
        return int(output_indexes[8:-2])
    elif type(output_indexes) is not str:
        return len(output_indexes)
    
def rec_fill_masks(mask_matrix, cluster_matrix, M, ind=None):
    if ind is None:
        ind = cluster_matrix.shape[0] - 1 + M
        
    if ind < M:
        mask_matrix[ind] = np.array([ind])
        return

    lind = int(cluster_matrix[ind-M,0])
    rind = int(cluster_matrix[ind-M,1])
    
    rec_fill_masks(mask_matrix, cluster_matrix, M, lind)
    mask_matrix[ind] = mask_matrix[lind]

    rec_fill_masks(mask_matrix, cluster_matrix, M, rind)
    mask_matrix[ind] = np.concatenate((mask_matrix[ind], mask_matrix[rind]))

# import scipy.sparse
# def make_masks(cluster_matrix):

#     # build the mask matrix recursively as an array of index lists
#     global count
#     count = 0
#     M = cluster_matrix.shape[0] + 1
#     mask_matrix_inds = np.zeros(2 * M - 1, dtype=np.object)
#     rec_fill_masks(mask_matrix_inds, cluster_matrix, M)
    
#     # convert the array of index lists into CSR format
#     indptr = np.zeros(len(mask_matrix_inds) + 1, dtype=np.int)
#     indices = np.zeros(np.sum([len(v) for v in mask_matrix_inds]), dtype=np.int)
#     pos = 0
#     for i in range(len(mask_matrix_inds)):
#         inds = mask_matrix_inds[i]
#         indices[pos:pos+len(inds)] = inds
#         pos += len(inds)
#         indptr[i+1] = pos
#     mask_matrix = scipy.sparse.csr_matrix(
#         (np.ones(len(indices), dtype=np.bool), indices, indptr),
#         shape=(len(mask_matrix_inds), M)
#     )

#     return mask_matrix



import pandas as pd
import numpy as np
import scipy as sp
import sys
import warnings
import copy
from slicer import Slicer, Alias
# from ._order import Order
from .utils._general import OpChain

# slicer confuses pylint...
# pylint: disable=no-member


op_chain_root = OpChain("shap.Explanation")
class MetaExplanation(type):
    """ This metaclass exposes the Explanation object's methods for creating template op chains.
    """

    def __getitem__(cls, item):
        return op_chain_root.__getitem__(item)

    @property
    def abs(cls):
        return op_chain_root.abs

    @property
    def argsort(cls):
        return op_chain_root.argsort

    @property
    def sum(cls):
        return op_chain_root.sum

    @property
    def max(cls):
        return op_chain_root.max

    @property
    def min(cls):
        return op_chain_root.min

    @property
    def mean(cls):
        return op_chain_root.mean
    
    @property
    def sample(cls):
        return op_chain_root.sample

    @property
    def hclust(cls):
        return op_chain_root.hclust


class Explanation(object, metaclass=MetaExplanation):
    """ This is currently an experimental feature don't depend on this object yet! :)
    """
    def __init__(
        self,
        values,
        base_values = None,
        data = None,
        display_data = None,
        instance_names = None,
        feature_names = None,
        output_names = None,
        output_indexes = None,
        lower_bounds = None,
        upper_bounds = None,
        main_effects = None,
        hierarchical_values = None,
        clustering = None
    ):
        self.transform_history = []

        # cloning. TODO: better cloning :)
        if issubclass(type(values), Explanation):
            e = values
            values = e.values
            base_values = e.base_values
            data = e.data
            
        output_dims = compute_output_dims(values, base_values, data)

        if len(_compute_shape(feature_names)) == 1: # TODO: should always be an alias once slicer supports per-row aliases
            values_shape = _compute_shape(values)
            if len(values_shape) >= 1 and len(feature_names) == values_shape[0]:
                feature_names = Alias(feature_names, 0)
            elif len(values_shape) >= 2 and len(feature_names) == values_shape[1]:
                feature_names = Alias(feature_names, 1)

        self._s = Slicer(
            values = values,
            base_values = base_values,
            data = data,
            display_data = display_data,
            instance_names = None if instance_names is None else Alias(instance_names, 0),
            feature_names = feature_names, 
            output_names = None if output_names is None else Alias(output_names, output_dims),
            output_indexes = None if output_indexes is None else (output_dims, output_indexes),
            lower_bounds = lower_bounds,
            upper_bounds = lower_bounds,
            main_effects = main_effects,
            hierarchical_values = hierarchical_values, #Obj(hierarchical_values, (0,None)),
            clustering = clustering
        )

    @property
    def shape(self):
        return _compute_shape(self._s.values)

    @property
    def values(self):
        return self._s.values
    @values.setter
    def values(self, new_values):
        self._s.values = new_values

    @property
    def base_values(self):
        return self._s.base_values
    @base_values.setter
    def base_values(self, new_base_values):
        self._s.base_values = new_base_values

    @property
    def data(self):
        return self._s.data
    @data.setter
    def data(self, new_data):
        self._s.data = new_data

    @property
    def display_data(self):
        return self._s.display_data
    @display_data.setter
    def display_data(self, new_display_data):
        self._s.display_data = new_display_data

    @property
    def instance_names(self):
        return self._s.instance_names

    @property
    def output_names(self):
        return self._s.output_names

    @property
    def output_indexes(self):
        return self._s.output_indexes

    @property
    def feature_names(self):
        return self._s.feature_names
    @feature_names.setter
    def feature_names(self, new_feature_names):
        self._s.feature_names = new_feature_names

    @property
    def lower_bounds(self):
        return self._s.lower_bounds

    @property
    def upper_bounds(self):
        return self._s.upper_bounds

    @property
    def main_effects(self):
        return self._s.main_effects
    @main_effects.setter
    def main_effects(self, new_main_effects):
        self._s.main_effects = new_main_effects

    @property
    def hierarchical_values(self):
        return self._s.hierarchical_values
    @hierarchical_values.setter
    def hierarchical_values(self, new_hierarchical_values):
        self._s.hierarchical_values = new_hierarchical_values

    @property
    def clustering(self):
        return self._s.clustering
    @clustering.setter
    def clustering(self, new_clustering):
        self._s.clustering = new_clustering
        
    def __repr__(self):
        out = ".values =\n"+self.values.__repr__()
        if self.base_values is not None:
            out += "\n\n.base_values =\n"+self.base_values.__repr__()
        if self.data is not None:
            out += "\n\n.data =\n"+self.data.__repr__()
        return out
    
    def __getitem__(self, item):
        """ This adds support for magic string indexes like "rank(0)".
        """
        if not isinstance(item, tuple):
            item = (item,)
        
        # convert any OpChains or magic strings
        for i,t in enumerate(item):
            orig_t = t
            if issubclass(type(t), OpChain):
                t = t.apply(self)
                if issubclass(type(t), (np.int64, np.int32)): # because slicer does not like numpy indexes
                    t = int(t)
                elif issubclass(type(t), np.ndarray):
                    t = [int(v) for v in t] # slicer wants lists not numpy arrays for indexing
            elif issubclass(type(t), Explanation):
                t = t.values
            elif type(t) is str:
                if is_1d(self.feature_names):
                    ind = np.where(np.array(self.feature_names) == t)[0][0]
                    t = int(ind)
                else:
                    new_values = []
                    new_data = []
                    for i in range(len(self.values)):
                        for s,v,d in zip(self.feature_names[i], self.values[i], self.data[i]):
                            if s == t:
                                new_values.append(v)
                                new_data.append(d)
                    new_self = copy.deepcopy(self)
                    new_self.values = new_values
                    new_self.data = new_data
                    new_self.feature_names = t
                    new_self.clustering = None
                    return new_self
            if issubclass(type(t), np.ndarray):
                t = [int(j) for j in t]
            elif issubclass(type(t), (np.int8, np.int16, np.int32, np.int64)):
                t = int(t)
            
            if t is not orig_t:
                tmp = list(item)
                tmp[i] = t
                item = tuple(tmp)
        
        # call slicer for the real work
        new_self = copy.copy(self)
        new_self.transform_history.append(("__getitem__", (item,)))
        new_self._s = self._s.__getitem__(item)
        
        return new_self

    def __len__(self):
        return self.shape[0]

    def __copy__(self):
        return Explanation(
            self.values,
            self.base_values,
            self.data,
            self.display_data,
            self.instance_names,
            self.feature_names,
            self.output_names,
            self.output_indexes,
            self.lower_bounds,
            self.upper_bounds,
            self.main_effects,
            self.hierarchical_values,
            self.clustering
        )

    @property
    def abs(self):
        new_self = copy.copy(self)
        new_self.values = np.abs(new_self.values)
        new_self.transform_history.append(("abs", None))
        return new_self

    def _numpy_func(self, fname, **kwargs):
        new_self = copy.copy(self)
        axis = kwargs.get("axis", None)

        # collapse the slicer to right shape
        if axis == 0:
            new_self = new_self[0]
        elif axis == 1:
            new_self = new_self[1]
        elif axis == 2:
            new_self = new_self[2]


        if self.feature_names is not None and not is_1d(self.feature_names) and axis == 0:
            new_values = self._flatten_feature_names()
            new_self.feature_names = np.array(list(new_values.keys()))
            new_self.values = np.array([getattr(np, fname)(v) for v in new_values.values()])
            new_self.clustering = None
        else:
            new_self.values = getattr(np, fname)(np.array(self.values), **kwargs)
            if new_self.data is not None:
                try:
                    new_self.data = getattr(np, fname)(np.array(self.data), **kwargs)
                except:
                    new_self.data = None
            if new_self.base_values is not None and issubclass(type(axis), int) and len(self.base_values.shape) > axis:
                new_self.base_values = getattr(np, fname)(self.base_values, **kwargs)
            elif issubclass(type(axis), int):
                new_self.base_values = None

        if axis == 0 and self.clustering is not None and len(self.clustering.shape) == 3:
            if self.clustering.std(0).sum() < 1e-8:
                new_self.clustering = self.clustering[0]
            else:
                new_self.clustering = None
        
        new_self.transform_history.append((fname, kwargs))
        
        return new_self

    def mean(self, axis):
        return self._numpy_func("mean", axis=axis)

    def max(self, axis):
        return self._numpy_func("max", axis=axis)

    def min(self, axis):
        return self._numpy_func("min", axis=axis)

    def sum(self, axis):
        return self._numpy_func("sum", axis=axis)

    @property
    def abs(self):
        return self._numpy_func("abs")

    @property
    def argsort(self):
        return self._numpy_func("argsort")

    @property
    def flip(self):
        return self._numpy_func("flip")


    def hclust(self, metric="sqeuclidean", axis=0):
        """ Computes an optimal leaf ordering sort order using hclustering.
        
        hclust(metric="sqeuclidean")
        
        Parameters
        ----------
        metric : string
            A metric supported by scipy clustering.

        axis : int
            The axis to cluster along.
        """
        values = self.values

        if len(values.shape) != 2:
            raise Exception("The hclust order only supports 2D arrays right now!")

        if axis == 1:
            values = values.T

        # compute a hierarchical clustering and return the optimal leaf ordering
        D = sp.spatial.distance.pdist(values, metric)
        cluster_matrix = sp.cluster.hierarchy.complete(D)
        inds = sp.cluster.hierarchy.leaves_list(sp.cluster.hierarchy.optimal_leaf_ordering(cluster_matrix, D))
        return inds

    def sample(self, max_samples, replace=False, random_state=0):
        """ Randomly samples the instances (rows) of the Explanation object.

        Parameters
        ----------
        max_samples : int
            The number of rows to sample. Note that if replace=False then less than
            fewer than max_samples will be drawn if explanation.shape[0] < max_samples.
        
        replace : bool
            Sample with or without replacement.
        """
        prev_seed = np.random.seed(random_state)
        inds = np.random.choice(self.shape[0], min(max_samples, self.shape[0]), replace=replace)
        np.random.seed(prev_seed)
        return self[list(inds)]

    def _flatten_feature_names(self):
        new_values = {}
        for i in range(len(self.values)):
            for s,v in zip(self.feature_names[i], self.values[i]):
                if s not in new_values:
                    new_values[s] = []
                new_values[s].append(v)
        return new_values

    def _use_data_as_feature_names(self):
        new_values = {}
        for i in range(len(self.values)):
            for s,v in zip(self.data[i], self.values[i]):
                if s not in new_values:
                    new_values[s] = []
                new_values[s].append(v)
        return new_values

    def percentile(self, q, axis=None):
        new_self = copy.deepcopy(self)
        if self.feature_names is not None and not is_1d(self.feature_names) and axis == 0:
            new_values = self._flatten_feature_names()
            new_self.feature_names = np.array(list(new_values.keys()))
            new_self.values = np.array([np.percentile(v, q) for v in new_values.values()])
            new_self.clustering = None
        else:
            new_self.values = np.percentile(new_self.values, q, axis)
            new_self.data = np.percentile(new_self.data, q, axis)
        #new_self.data = None
        new_self.transform_history.append(("percentile", (axis,)))
        return new_self

def compute_output_dims(values, base_values, data):
    values_shape = _compute_shape(values)

    # input shape matches the data shape
    if data is not None:
        data_shape = _compute_shape(data)

    # if we are not given any data we assume it would be the same shape as the given values
    else: 
        data_shape = values_shape

    # output shape is known from the base values
    if base_values is not None:
        output_shape = _compute_shape(base_values)[1:]
    else:
        output_shape = tuple()

    interaction_order = len(values_shape) - len(data_shape) - len(output_shape)
    values_dims = list(range(len(values_shape)))
    output_dims = range(len(data_shape) + interaction_order, values_dims[-1])
    return tuple(output_dims)

def is_1d(val):
    return not (isinstance(val[0], list) or isinstance(val[0], np.ndarray))

class Op():
    pass

class Percentile(Op):
    def __init__(self, percentile):
        self.percentile = percentile

    def add_repr(self, s, verbose=False):
        return "percentile("+s+", "+str(self.percentile)+")"

    


def _compute_shape(x):
    if not hasattr(x, "__len__"):
        return tuple()
    elif len(x) > 0 and isinstance(x[0], str):
        return (None,)
    else:
        if type(x) == dict:
            return (len(x),) + _compute_shape(x[next(iter(x))])

        # 2D arrays we just take their shape as-is
        if len(getattr(x, "shape", tuple())) > 1:
            return x.shape

        # 1D arrays we need to look inside
        if len(x) == 0:
            return (0,)
        elif len(x) == 1:
            return (len(x),) + _compute_shape(x[0])
        else:
            first_shape = _compute_shape(x[0])
            for i in range(1,len(x)):
                shape = _compute_shape(x[i])
                if shape != first_shape:
                    return (len(x), None)
            return (len(x),) + first_shape


import warnings
try:
    import matplotlib.pyplot as pl
except ImportError:
    warnings.warn("matplotlib could not be loaded!")
    pass
from ._labels import labels
from ..utils import format_value, ordinal_str
from ._utils import convert_ordering, convert_color
from . import colors
import numpy as np
import scipy
import copy
from .. import Explanation


# TODO: improve the bar chart to look better like the waterfall plot with numbers inside the bars when they fit
def bar(shap_values, max_display=10, order=Explanation.abs, clustering=None, cluster_threshold=0.5, show=True):
    """ Create a bar plot of a set of SHAP values.

    If a single sample is passed then we plot the SHAP values as a bar chart. If an
    Explanation with many samples is passed then we plot the mean absolute value for
    each feature column as a bar chart.


    Parameters
    ----------
    shap_values : shap.Explanation
        A single row of a SHAP Explanation object (i.e. shap_values[0]) or a multi-row Explanation
        object that we want to summarize.

    max_display : int
        The maximum number of bars to display.

    show : bool
        If show is set to False then we don't call the matplotlib.pyplot.show() function. This allows
        further customization of the plot by the caller after the bar() function is finished. 

    """

    assert str(type(shap_values)).endswith("Explanation'>"), "The shap_values paramemter must be a shap.Explanation object!"

    if len(shap_values.shape) == 2:
        shap_values = shap_values.abs.mean(0)

    # unpack the Explanation object
    features = shap_values.data
    feature_names = shap_values.feature_names
    if clustering is None:
        partition_tree = getattr(shap_values, "clustering", None)
    elif clustering is False:
        partition_tree = None
    else:
        partition_tree = clustering
    if partition_tree is not None:
        assert partition_tree.shape[1] == 4, "The clustering provided by the Explanation object does not seem to be a partition tree (which is all shap.plots.bar supports)!"
    transform_history = shap_values.transform_history
    values = np.array(shap_values.values)

    if len(values) == 0:
        raise Exception("The passed Explanation is empty! (so there is nothing to plot)")

    # TODO: Rather than just show the "1st token", "2nd token", etc. it would be better to show the "Instance 0's 1st but", etc
    if issubclass(type(feature_names), str):
        feature_names = [ordinal_str(i)+" "+feature_names for i in range(len(values))]

    # build our auto xlabel based on the transform history of the Explanation object
    xlabel = "SHAP value"
    for op in transform_history:
        if op[0] == "abs":
            xlabel = "|"+xlabel+"|"
        elif op[0] == "__getitem__":
            pass # no need for slicing to effect our label
        else:
            xlabel = str(op[0])+"("+xlabel+")"

    # unwrap any pandas series
    if str(type(features)) == "<class 'pandas.core.series.Series'>":
        if feature_names is None:
            feature_names = list(features.index)
        features = features.values
    
    # ensure we at least have default feature names
    if feature_names is None:
        feature_names = np.array([labels['FEATURE'] % str(i) for i in range(len(values))])
    
    # determine how many top features we will plot
    if max_display is None:
        max_display = len(feature_names)
    num_features = min(max_display, len(values))

    # iteratively merge nodes until we can cut off the smallest feature values to stay within
    # num_features without breaking a cluster tree
    orig_inds = [[i] for i in range(len(values))]
    orig_values = values.copy()
    while True:
        feature_order = convert_ordering(order, Explanation(values))
        if partition_tree is not None:

            # compute the leaf order if we were to show (and so have the ordering respect) the whole partition tree
            clust_order = sort_inds(partition_tree, np.abs(values))

            # now relax the requirement to match the parition tree ordering for connections above cluster_threshold
            dist = scipy.spatial.distance.squareform(scipy.cluster.hierarchy.cophenet(partition_tree))
            feature_order = get_sort_order(values, dist, clust_order, cluster_threshold, feature_order)
        
            # if the last feature we can display is connected in a tree the next feature then we can't just cut
            # off the feature ordering, so we need to merge some tree nodes and then try again.
            if max_display < len(feature_order) and dist[feature_order[max_display-1],feature_order[max_display-2]] <= cluster_threshold:
                values, partition_tree, orig_inds = merge_nodes(values, partition_tree, orig_inds)
            else:
                break
        else:
            break
    
    # here we build our feature names, accounting for the fact that some features might be merged together
    feature_inds = feature_order[:max_display]
    y_pos = np.arange(len(feature_inds), 0, -1)
    feature_names_new = []
    for pos,inds in enumerate(orig_inds):
        if len(inds) == 1:
            feature_names_new.append(feature_names[inds[0]])
        elif len(inds) <= 2:
            feature_names_new.append(" + ".join([feature_names[i] for i in inds]))
        else:
            max_ind = np.argmax(orig_values[inds])
            feature_names_new.append(feature_names[inds[max_ind]] + " + %d other features" % (len(inds)-1))
    feature_names = feature_names_new

    # see how many individual (vs. grouped at the end) features we are plotting
    if num_features < len(values):
        num_cut = np.sum([len(orig_inds[feature_order[i]]) for i in range(num_features-1, len(values))])
        values[feature_order[num_features-1]] = np.sum([values[feature_order[i]] for i in range(num_features-1, len(values))])
    
    # build our y-tick labels
    yticklabels = []
    for i in feature_inds:
        if features is not None:
            yticklabels.append(format_value(features[i], "%0.03f") + " = " + feature_names[i])
        else:
            yticklabels.append(feature_names[i])
    if num_features < len(values):
        yticklabels[-1] = "Sum of %d other features" % num_cut

    # compute our figure size based on how many features we are showing
    row_height = 0.5
    pl.gcf().set_size_inches(8, num_features * row_height + 1.5)

    # if negative values are present then we draw a vertical line to mark 0, otherwise the axis does this for us...
    negative_values_present = np.sum(values[feature_order[:num_features]] < 0) > 0
    if negative_values_present:
        pl.axvline(0, 0, 1, color="#000000", linestyle="-", linewidth=1, zorder=1)

    # draw the bars
    pl.barh(
        y_pos, values[feature_inds],
        0.7, align='center',
        color=[colors.blue_rgb if values[feature_inds[i]] <= 0 else colors.red_rgb for i in range(len(y_pos))]
    )

    # draw the yticks
    pl.yticks(list(y_pos) + list(y_pos), yticklabels + [l.split('=')[-1] for l in yticklabels], fontsize=13)

    xlen = pl.xlim()[1] - pl.xlim()[0]
    fig = pl.gcf()
    ax = pl.gca()
    #xticks = ax.get_xticks()
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    bbox_to_xscale = xlen/width

    for i in range(len(y_pos)):
        ind = feature_order[i]
        if values[ind] < 0:
            pl.text(
                values[ind] - (5/72)*bbox_to_xscale, y_pos[i], format_value(values[ind], '%+0.02f'),
                horizontalalignment='right', verticalalignment='center', color=colors.blue_rgb,
                fontsize=12
            )
        else:
            pl.text(
                values[ind] + (5/72)*bbox_to_xscale, y_pos[i], format_value(values[ind], '%+0.02f'),
                horizontalalignment='left', verticalalignment='center', color=colors.red_rgb,
                fontsize=12
            )

    # put horizontal lines for each feature row
    for i in range(num_features):
        pl.axhline(i+1, color="#888888", lw=0.5, dashes=(1, 5), zorder=-1)
    
    if features is not None:
        features = list(features)

        # try and round off any trailing zeros after the decimal point in the feature values
        for i in range(len(features)):
            try:
                if round(features[i]) == features[i]:
                    features[i] = int(features[i])
            except TypeError:
                pass # features[i] must not be a number
    
    pl.gca().xaxis.set_ticks_position('bottom')
    pl.gca().yaxis.set_ticks_position('none')
    pl.gca().spines['right'].set_visible(False)
    pl.gca().spines['top'].set_visible(False)
    if negative_values_present:
        pl.gca().spines['left'].set_visible(False)
    pl.gca().tick_params('x', labelsize=11)

    xmin,xmax = pl.gca().get_xlim()
    
    if negative_values_present:
        pl.gca().set_xlim(xmin - (xmax-xmin)*0.05, xmax + (xmax-xmin)*0.05)
    else:
        pl.gca().set_xlim(xmin, xmax + (xmax-xmin)*0.05)
    
    # if features is None:
    #     pl.xlabel(labels["GLOBAL_VALUE"], fontsize=13)
    # else:
    pl.xlabel(xlabel, fontsize=13)

    # color the y tick labels that have the feature values as gray
    # (these fall behind the black ones with just the feature name)
    tick_labels = pl.gca().yaxis.get_majorticklabels()
    for i in range(num_features):
        tick_labels[i].set_color("#999999")

    # draw a dendrogram if we are given a partition tree
    if partition_tree is not None:
        
        # compute the dendrogram line positions based on our current feature order
        feature_pos = np.argsort(feature_order)
        ylines,xlines = dendrogram_coords(feature_pos, partition_tree)
        
        # plot the distance cut line above which we don't show tree edges
        xmin,xmax = pl.xlim()
        xlines_min,xlines_max = np.min(xlines),np.max(xlines)
        l = pl.axvline((cluster_threshold / (xlines_max - xlines_min)) * 0.1 * (xmax - xmin) + xmax, color="#dddddd", dashes=(1, 1))
        l.set_clip_on(False)
        
        for (xline, yline) in zip(xlines, ylines):
            
            # normalize the x values to fall between 0 and 1
            xv = (np.array(xline) / (xlines_max - xlines_min))

            # only draw if we are not going past distance threshold
            if np.array(xline).max() <= cluster_threshold:

                # only draw if we are not going past the bottom of the plot
                if yline.max() < max_display:
                    l = pl.plot(
                        xv * 0.1 * (xmax - xmin) + xmax,
                        max_display - np.array(yline),
                        color="#999999"
                    )
                    for v in l:
                        v.set_clip_on(False)
    
    if show:
        pl.show()

def get_sort_order(shap_values, dist, clust_order, cluster_threshold, feature_order):
    """ Returns a sorted order of the values where we respect the clustering order when dist[i,j] < cluster_threshold
    """
    
    #feature_imp = np.abs(values)

    # if partition_tree is not None:
    #     new_tree = fill_internal_max_values(partition_tree, shap_values)
    #     clust_order = sort_inds(new_tree, np.abs(shap_values))
    clust_inds = np.argsort(clust_order)

    feature_order = feature_order.copy()#order.apply(Explanation(shap_values))
    # print("feature_order", feature_order)
    for i in range(len(feature_order)-1):
        ind1 = feature_order[i]
        next_ind = feature_order[i+1]
        next_ind_pos = i + 1
        for j in range(i+1,len(feature_order)):
            ind2 = feature_order[j]

            
            
            #if feature_imp[ind] > 
            # if ind1 == 2:
            #     print(ind1, ind2, dist[ind1,ind2])
            if dist[ind1,ind2] <= cluster_threshold:
                
                # if ind1 == 2:
                #     print(clust_inds)
                #     print(ind1, ind2, next_ind, dist[ind1,ind2], clust_inds[ind2], clust_inds[next_ind])
                if dist[ind1,next_ind] > cluster_threshold or clust_inds[ind2] < clust_inds[next_ind]:
                    next_ind = ind2
                    next_ind_pos = j
            # print("next_ind", next_ind)
            # print("next_ind_pos", next_ind_pos)
        
        # insert the next_ind next
        for j in range(next_ind_pos, i+1, -1):
            #print("j", j)
            feature_order[j] = feature_order[j-1]
        feature_order[i+1] = next_ind
        #print(feature_order)

        
    
    return feature_order

def merge_nodes(values, partition_tree, orig_inds):
    """ This merges the two clustered leaf nodes with the smallest total value.
    """
    M = partition_tree.shape[0] + 1

    ptind = 0
    min_val = np.inf
    for i in range(partition_tree.shape[0]):
        ind1 = int(partition_tree[i,0])
        ind2 = int(partition_tree[i,1])
        if ind1 < M and ind2 < M:
            val = np.abs(values[ind1]) + np.abs(values[ind2])
            if val < min_val:
                min_val = val
                ptind = i
                #print("ptind", ptind, min_val)

    ind1 = int(partition_tree[ptind,0])
    ind2 = int(partition_tree[ptind,1])
    if ind1 > ind2:
        tmp = ind1
        ind1 = ind2
        ind2 = tmp
    
    values_new = values.copy()
    values_new[ind1] += values_new[ind2]
    values_new = np.delete(values_new, ind2)
    
    partition_tree_new = partition_tree.copy()
    for i in range(partition_tree_new.shape[0]):
        i0 = int(partition_tree_new[i,0])
        i1 = int(partition_tree_new[i,1])
        if i0 == ind2:
            partition_tree_new[i,0] = ind1
        elif i0 > ind2:
            partition_tree_new[i,0] -= 1
            if i0 == ptind + M:
                partition_tree_new[i,0] = ind1
            elif i0 > ptind + M:
                partition_tree_new[i,0] -= 1

            
        if i1 == ind2:
            partition_tree_new[i,1] = ind1
        elif i1 > ind2:
            partition_tree_new[i,1] -= 1
            if i1 == ptind + M:
                partition_tree_new[i,1] = ind1
            elif i1 > ptind + M:
                partition_tree_new[i,1] -= 1
    partition_tree_new = np.delete(partition_tree_new, ptind, axis=0)
    
    orig_inds_new = copy.deepcopy(orig_inds)
    orig_inds_new[ind1] += orig_inds_new[ind2]
    del orig_inds_new[ind2]

    # update the counts to be correct
    fill_counts(partition_tree_new)
    
    return values_new, partition_tree_new, orig_inds_new

def dendrogram_coords(leaf_positions, partition_tree):
    """ Returns the x and y coords of the lines of a dendrogram where the leaf order is given.

    Note that scipy can compute these coords as well, but it does not allow you to easily specify
    a specific leaf order, hence this reimplementation.
    """
    
    xout = []
    yout = []
    _dendrogram_coords_rec(partition_tree.shape[0]-1, leaf_positions, partition_tree, xout, yout)
    
    return np.array(xout), np.array(yout)
def _dendrogram_coords_rec(pos, leaf_positions, partition_tree, xout, yout):
    M = partition_tree.shape[0] + 1
    
    if pos < 0:
        return leaf_positions[pos + M], 0
    
    left = int(partition_tree[pos, 0]) - M
    right = int(partition_tree[pos, 1]) - M
    
    x_left, y_left = _dendrogram_coords_rec(left, leaf_positions, partition_tree, xout, yout)
    x_right, y_right = _dendrogram_coords_rec(right, leaf_positions, partition_tree, xout, yout)
    
    y_curr = partition_tree[pos, 2]
    
    xout.append([x_left, x_left, x_right, x_right])
    yout.append([y_left, y_curr, y_curr, y_right])
    
    return (x_left + x_right) / 2, y_curr

def fill_internal_max_values(partition_tree, leaf_values):
    """ This fills the forth column of the partition tree matrix with the max leaf value in that cluster.
    """
    M = partition_tree.shape[0] + 1
    new_tree = partition_tree.copy()
    for i in range(new_tree.shape[0]):
        val = 0
        if new_tree[i,0] < M:
            ind = int(new_tree[i,0])
            val = max(val, np.abs(leaf_values[ind]))
        else:
            ind = int(new_tree[i,0])-M
            val = max(val, np.abs(new_tree[ind,3])) # / partition_tree[ind,2])
        if new_tree[i,1] < M:
            ind = int(new_tree[i,1])
            val = max(val, np.abs(leaf_values[ind]))
        else:
            ind = int(new_tree[i,1])-M
            val = max(val, np.abs(new_tree[ind,3])) # / partition_tree[ind,2])
        new_tree[i,3] = val
    return new_tree

def fill_counts(partition_tree):
    """ This updates the 
    """
    M = partition_tree.shape[0] + 1
    for i in range(partition_tree.shape[0]):
        val = 0
        if partition_tree[i,0] < M:
            ind = int(partition_tree[i,0])
            val += 1
        else:
            ind = int(partition_tree[i,0])-M
            val += partition_tree[ind,3]
        if partition_tree[i,1] < M:
            ind = int(partition_tree[i,1])
            val += 1
        else:
            ind = int(partition_tree[i,1])-M
            val += partition_tree[ind,3]
        partition_tree[i,3] = val

def sort_inds(partition_tree, leaf_values, pos=None, inds=None):
    if inds is None:
        inds = []
    
    if pos is None:
        partition_tree = fill_internal_max_values(partition_tree, leaf_values)
        pos = partition_tree.shape[0]-1
    
    M = partition_tree.shape[0] + 1
        
    if pos < 0:
        inds.append(pos + M)
        return
    
    left = int(partition_tree[pos, 0]) - M
    right = int(partition_tree[pos, 1]) - M
    
    
    left_val = partition_tree[left,3] if left >= 0 else leaf_values[left + M]
    right_val = partition_tree[right,3] if right >= 0 else leaf_values[right + M]
    
    if left_val < right_val:
        tmp = right
        right = left
        left = tmp
    
    sort_inds(partition_tree, leaf_values, left, inds)
    sort_inds(partition_tree, leaf_values, right, inds)
    
    return inds

# def compute_sort_counts(partition_tree, leaf_values, pos=None):
#     if pos is None:
#         pos = partition_tree.shape[0]-1
    
#     M = partition_tree.shape[0] + 1
        
#     if pos < 0:
#         return 1,leaf_values[pos + M]
    
#     left = int(partition_tree[pos, 0]) - M
#     right = int(partition_tree[pos, 1]) - M
    
#     left_val,left_sum = compute_sort_counts(partition_tree, leaf_values, left)
#     right_val,right_sum = compute_sort_counts(partition_tree, leaf_values, right)
    
#     if left_sum > right_sum:
#         left_val = right_val + 1
#     else:
#         right_val = left_val + 1

#     if left >= 0:
#         partition_tree[left,3] = left_val
#     if right >= 0:
#         partition_tree[right,3] = right_val

    
#     return max(left_val, right_val) + 1, max(left_sum, right_sum)

def bar_legacy(shap_values, features=None, feature_names=None, max_display=None, show=True):
    
    # unwrap pandas series
    if str(type(features)) == "<class 'pandas.core.series.Series'>":
        if feature_names is None:
            feature_names = list(features.index)
        features = features.values
        
    if feature_names is None:
        feature_names = np.array([labels['FEATURE'] % str(i) for i in range(len(shap_values))])
    
    if max_display is None:
        max_display = 7
    else:
        max_display = min(len(feature_names), max_display)
        
    
    feature_order = np.argsort(-np.abs(shap_values))
    
    # 
    feature_inds = feature_order[:max_display]
    y_pos = np.arange(len(feature_inds), 0, -1)
    pl.barh(
        y_pos, shap_values[feature_inds],
        0.7, align='center',
        color=[colors.red_rgb if shap_values[feature_inds[i]] > 0 else colors.blue_rgb for i in range(len(y_pos))]
    )
    pl.yticks(y_pos, fontsize=13)
    if features is not None:
        features = list(features)

        # try and round off any trailing zeros after the decimal point in the feature values
        for i in range(len(features)):
            try:
                if round(features[i]) == features[i]:
                    features[i] = int(features[i])
            except TypeError:
                pass # features[i] must not be a number
    yticklabels = []
    for i in feature_inds:
        if features is not None:
            yticklabels.append(feature_names[i] + " = " + str(features[i]))
        else:
            yticklabels.append(feature_names[i])
    pl.gca().set_yticklabels(yticklabels)
    pl.gca().xaxis.set_ticks_position('bottom')
    pl.gca().yaxis.set_ticks_position('none')
    pl.gca().spines['right'].set_visible(False)
    pl.gca().spines['top'].set_visible(False)
    #pl.gca().spines['left'].set_visible(False)
    
    pl.xlabel("SHAP value (impact on model output)")
    
    if show:
        pl.show()


from .. import Explanation
from ..utils import OpChain
from . import colors

def convert_color(color):
    try:
        color = pl.get_cmap(color)
    except:
        pass
    
    if color == "shap_red":
        color = colors.red_rgb
    elif color == "shap_blue":
        color = colors.blue_rgb
    
    return color

def convert_ordering(ordering, shap_values):
    if issubclass(type(ordering), OpChain):
        ordering = ordering.apply(Explanation(shap_values))
    if issubclass(type(ordering), Explanation):
        if "argsort" in [n for n,a in ordering.transform_history]:
            ordering = ordering.values
        else:
            ordering = ordering.argsort.flip.values
    return ordering